import sys
from importlib import util
from copy import deepcopy
from os import path, listdir

from .socket import SendSocket
from .command import Command, CommandResponse
from .message import *
from .errors import ExtensionAlreadyLoaded, ExtensionNotFound, ExtensionFailed, NoEntryPointError 
from .log import Log


class Channel():

    name: str
    __modules: dict
    __commands: dict
    socket: SendSocket

    def __init__(self, name: str = "undefined"):
        self.name = name
        self.__modules = {}
        self.__commands = {}

        self.log = Log(instance=name, client="Twitch Channel")
        pass

    # Public commands for use in the Bridge and in Commands

    def propogate(self, socket: SendSocket, channel_name: str):
        self.name = channel_name
        self.log.set_instance(channel_name)
        newchannel = deepcopy(self)
        newchannel.__cache_commands()
        newchannel.socket = socket
        if channel_name != "undefined":
            newchannel.socket.join(channel_name)
        return newchannel
        


    def send(self, message):
        self.log.debug(f"not implemented: SEND > {message}")
        # self.__bridge..... 
        pass

    def load_command(self, command):
        self.__commands[command.name] = command



    # Callbacks used by the bridge for all Messages sent via the inbound socket.

    ### to do: add the rest of the message types in here

    # Base Callback
    def on_raw(self, data: RawMessage):
        self.log.debug("`on_raw` invoked.")
        for command in self.__get_command_with_handler("on_raw"):
            command.on_raw(data)

    # IRC: Membership Callbacks

    def on_join(self, data: Join):
        self.log.debug("`on_join` invoked.")
        for command in self.__get_command_with_handler("on_join"):
            command.on_join(data)

    def on_part(self, data: Part):
        self.log.debug("`on_part` invoked.")
        for command in self.__get_command_with_handler("on_part"):
            command.on_part(data)

    def on_mode(self, data: Mode):
        self.log.debug("`on_mode` invoked.") 
        for command in self.__get_command_with_handler("on_mode"):
            command.on_mode(data)

    def on_names(self, data: Names):
        self.log.debug("`on_names` invoked.")
        for command in self.__get_command_with_handler("on_names"):
            command.on_names(data)

    # IRC: Command & Tag Callbacks

    def on_clearchat(self, data: ClearChat):
        self.log.debug("`on_clearchat` invoked.")
        for command in self.__get_command_with_handler("on_clearchat"):
            command.on_clearchat(data)

    def on_clearmessage(self, data: ClearMessage):
        self.log.debug("`on_clearmessage` invoked.")
        for command in self.__get_command_with_handler("on_clearmessage"):
            command.on_clearmessage(data)
    
    def on_hosttarget(self, data: HostTarget):
        self.log.debug("`on_hosttarget` invoked.")
        for command in self.__get_command_with_handler("on_hosttarget"):
            command.on_hosttarget(data)

    def on_notice(self, data: Notice):
        self.log.debug(f"`on_notice` invoked. (not implemented)")
        for command in self.__get_command_with_handler("on_notice"):
            command.on_notice(data)

    def on_reconnect(self, data: Reconnect):
        self.log.debug("`on_reconnect` invoked.")
        for command in self.__get_command_with_handler("on_reconnect"):
            command.on_reconnect(data)

    def on_roomstate(self, data: RoomState):
        self.log.debug("`on_roomstate` invoked.")
        for command in self.__get_command_with_handler("on_roomstate"):
            command.on_roomstate(data)

    def on_userstate(self, data: UserState):
        self.log.debug("`on_userstate` invoked.")
        for command in self.__get_command_with_handler("on_userstate"):
            command.on_userstate(data)

    def on_globaluserstate(self, data: GlobalUserState):
        self.log.debug("`on_globaluserstate` invoked.")
        for command in self.__get_command_with_handler("on_globaluserstate"):
            command.on_globaluserstate(data)

    def on_privmessage(self, data: PrivateMessage):
        self.log.debug("`on_privmessage` invoked.")
        for command in self.__get_command_with_handler("on_privmessage"):
            command.on_privmessage(data)
        
    def on_usernotice(self, data: UserNotice):
        self.log.debug("`on_usernotice` invoked.")
        for command in self.__get_command_with_handler("on_usernotice"):
            command.on_usernotice(data)

    def on_ritual_usernotice(self, data: RitualUserNotice):
        self.log.debug("`on_ritual_usernotice` invoked.")
        for command in self.__get_command_with_handler("on_ritual_usernotice"):
            command.on_ritual_usernotice(data)

    def on_bitbadgeupgrade_usernotice(self, data: BitBadgeUpgradeUserNotice):
        self.log.debug("`on_bitbadgeupgrade_usernotice` invoked.")
        for command in self.__get_command_with_handler("on_bitbadgeupgrade_usernotice"):
            command.on_bitbadgeupgrade_usernotice(data)

    def on_raid_usernotice(self, data: RaidUserNotice):
        self.log.debug("`on_raid_usernotice` invoked.")
        for command in self.__get_command_with_handler("on_raid_usernotice"):
            command.on_raid_usernotice(data)

    def on_subscriber_usernotice(self, data: SubscriberUserNotice):
        self.log.debug("`on_subscriber_usernotice` invoked.")
        for command in self.__get_command_with_handler("on_subscriber_usernotice"):
            command.on_subscriber_usernotice(data)
    
    def on_giftedsubscriber_usernotice(self, data: GiftedSubscriberUserNotice):
        self.log.debug("`on_giftedsubscriber_usernotice` invoked.")
        for command in self.__get_command_with_handler("on_giftedsubscriber_usernotice"):
            command.on_giftedsubscriber_usernotice(data)

    def on_command(self, data: CommandMessage):
        self.log.debug("`on_command` invoked.")
        for command in self.__get_command_with_handler("on_command"):
            command.on_command(data)






    # backend commands for managing the loading and processing of command modules
    
    def __cache_commands(self):
        commands = []
        for _file in listdir(path.join(path.dirname(__file__), 'commands/')):
            if "__init__" not in _file:
                print ("r1: Found: ", _file)
                filename, ext = path.splitext(_file)
                if '.py' in ext:
                    commands.append(f'jarviscore.commands.{filename}')
        
        if path.exists("commands/"):
            for _file in listdir('commands/'):
                print ("r2: Found: ", _file)
                filename, ext = path.splitext(_file)
                if '.py' in ext:
                    commands.append(f'commands.{filename}')

        for extension in reversed(commands):
            try:
                print("attempting to load: ", extension)
                self._load_command(f'{extension}')
            except Exception as e:
                exc = f'{type(e).__name__}: {e}'
                print(f'Failed to load extension {extension}\n{exc}')
        

    def _load_command(self, name):
        if name in self.__modules:
            raise ExtensionAlreadyLoaded(name)

        spec = util.find_spec(name)
        if spec is None:
            raise ExtensionNotFound(name)

        self._load_from_module_spec(spec, name)

    
    def _load_from_module_spec(self, spec, key):
        lib = util.module_from_spec(spec)
        sys.modules[key] = lib
        try:
            spec.loader.exec_module(lib)
        except Exception as e:
            del sys.modules[key]
            raise ExtensionFailed(key, e) from e

        try:
            setup = getattr(lib, 'setup')
        except AttributeError:
            del sys.modules[key]
            raise NoEntryPointError(key)

        try:
            setup(self)
        except Exception as e:
            del sys.modules[key]
            self._call_module_finalizers(lib, key)
            raise ExtensionFailed(key, e) from e
        else:
            self.__modules[key] = lib

    

    def _call_module_finalizers(self, lib, key):
        try:
            teardown = getattr(lib, 'teardown')
        except AttributeError:
            pass
        else:
            try:
                teardown(self)
            except Exception:
                pass
        finally:
            self.__modules.pop(key, None)
            sys.modules.pop(key, None)

    def __get_command_with_handler(self, handler: str):
        for command in self.__commands:
            try:
                if hasattr(self.__commands[command], handler):
                    yield self.__commands[command]
            except AttributeError:
                pass
            


            


        
        

    



import pyodbc 
import pymysql
from .settings import Settings
from .helpers import Helpers
from datetime import datetime, timedelta





class _Model():
    DBConnection = None
    DBCursor = None

    def __execute(self, query):
        self.DBCursor.execute(query)

    def insert(self, query):
        """Insert data into the database"""
        self.DBCursor.execute(query)
        self.DBConnection.commit()

    def truncate(self, query):
        """Truncate a table in the database"""
        self.DBCursor.execute(query)
        self.DBConnection.commit()

    def update(self, query):
        """Insert data into the database"""
        self.DBCursor.execute(query)
        self.DBConnection.commit()
    
    def delete(self, query):
        """Delete data from the database"""
        self.DBCursor.execute(query)
        self.DBConnection.commit()

    def fetchOne(self, query):
        """Return the first record from the query"""
        self.__execute(query)
        return self.DBCursor.fetchone()
    
    def fetchAll(self, query):
        """Return all records from the query"""
        self.__execute(query)
        return self.DBCursor.fetchall()

    def get_timestamp_from_datetime(self, timestamp: datetime):
        """return a formatted date string for a given timestamp"""
        return Helpers().get_timestamp_from_datetime(timestamp)
        
    def get_timestamp(self):
        return Helpers().get_timestamp()
    
    def get_utc_timestamp(self):
        return Helpers().get_utc_timestamp()

    def utc_from_timestamp(self, utc_timestamp):
        return Helpers().utc_from_timestamp(utc_timestamp)

    def utc_from_timestamp_to_localtime_timestamp(self, utc_timestamp: str):
        return Helpers().utc_from_timestamp_to_localtime_timestamp(utc_timestamp)
    
    def sanitise(self, target):
        """Sanitise the input from undesirable data"""
        return target.replace("'", "''").replace("\"", "\"\"").replace(";", "")


class MSSQLModel(_Model):
    """A Model abstraction for interacting with Microsoft SQL Server."""
    def __init__(self):
        setting = Settings()
        self.DBConnection = pyodbc.connect(setting.get_setting("database.string"))
        self.DBCursor = self.DBConnection.cursor()

class MySQLModel(_Model):
    """A Model abstraction for interacting with MySQL."""
    def __init__(self, user, password, database="jarvis", host="127.0.0.1"):
        self.DBConnection = pymysql.connect(host=host, 
                                    charset='utf8mb4',
                                    user=user, 
                                    password=password, 
                                    db=database)
        self.DBCursor = self.DBConnection.cursor()


class Model(MySQLModel):
    """Default Model Class for Jarvis, preconfigured"""
    def __init__(self, user, password):
        MySQLModel.__init__(self, user, password)


from .channel import Channel
from .socket import SendSocket, ReadSocket
from .command import Command, CommandResponse
from .message import RawMessage
from time import sleep  

class Bridge(object):
    """description of class"""
    send_socket: SendSocket
    read_socket: ReadSocket
    channels: list
    active: bool
    


    def __init__(self, client):
        self.client = client
        self.active = True
        self.__init_sockets()
        self.__init_channels()
        pass


    def __init_sockets(self):
        self.read_socket = ReadSocket(self.client.nick, self.client.token)
        self.send_socket = SendSocket(self.client.nick, self.client.token)
        pass


    def __init_channels(self):
        self.channels = []
        proto_channel = Channel(name="proto")
        for chn in self.client.channels:
            self.channels.append(proto_channel.propogate(self.send_socket, chn))

    def __get_channel(self, name: str):
        for channel in self.channels:
            if channel.name == name:
                return channel
        

    def run(self):
        self.read_socket.activate(self.channels)
        self.send_socket.activate(self.channels)
        self.read_socket.start()
        self.send_socket.start()
        try:
            while self.active:
                sleep(0.0001)
        except KeyboardInterrupt:            
            self.read_socket.close()
            self.send_socket.close()


    def check_messages(self):
        pass
        



    def add_channel(self, channel: str):
        chn = Channel()
        self.channels.append(chn.propogate(self.send_socket, chn))
        self.send_socket.connect_to_channel(channel)
        self.read_socket.connect_to_channel(channel)
    
    def remove_channel(self, channel: str):
        for chn in self.channels:
            if chn.name == channel:
                self.channels.remove(chn)
        self.send_socket.disconnect_from_channel(channel)
        self.read_socket.disconnect_from_channel(channel)

    def leave_channel(self, channel: str):
        pass




    





from ..command import Command
from ..message import PrivateMessage, CommandMessage

class Ping(Command):

    def __init__(self, channel):
        Command.__init__(self, "Ping")
        self.channel = channel

    def on_privmessage(self, data: PrivateMessage):
        if "ping" == data.message_text and data.user_id == "82504138":
            self.channel.socket.send_message(data.channel, "Pong")
            log(f"sent, 'Pong' to '{data.channel}'")
    
    def on_command(self, data: CommandMessage):
        if "ping" == data.message_text and data.user_id == "82504138":
            self.channel.socket.send_message(data.channel, "Pong")
            log(f"sent, 'Pong' to '{data.channel}'")

def log(msg):
    print(f"CMD|PING: {msg}")


def setup(channel):
    channel.load_command(Ping(channel))
    log(f"[{channel.name}]: Loaded Module Ping")

def teardown(channel):
    log(f"[{channel.name}]: Removed Module Ping")


import socket
from time import sleep


from .log import Log
# from .bridge import Bridge
from threading import Thread
from .message import RawMessage


from .messageparser import parse_line


class Socket(Thread):
    
    # bridge: Bridge
    socket: socket.socket
    buffer: str
    active: bool
    ready: bool
    channel_list: list


    # def __init__(self, bridge: Bridge = None):
    def __init__(self, nick: str, token: str):
        # self.bridge = bridge
        Thread.__init__(self)
        self.nick = nick
        self.token = token
        self.active = True
        self.buffer = ""
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.channel_list = []
        

    def __connect(self):
        self.socket.connect(("irc.chat.twitch.tv", 6667))
        self._send_raw(f"PASS {self.token}")
        self._send_raw(f"NICK {self.nick}")
        self._send_raw("CAP REQ :twitch.tv/membership")
        self._send_raw("CAP REQ :twitch.tv/tags")
        self._send_raw("CAP REQ :twitch.tv/commands")
        pass

    def disconnect(self):
        print("departing channels")
        for channel in self.channel_list:
            self._send_raw("PART #"+ channel.name.lower() +" ")
        try:
            self.socket.close()
        except Exception:
            pass

    def reconnect(self):
        print("Reconnect detected!")
        self.disconnect()
        print("Waiting to reconnect.")
        sleep(10)
        self.activate(self.channel_list)

    def connect_to_channel(self, channel):
        self._send_raw("JOIN #"+ channel.name.lower() +" ")
        self.channel_list.append(channel)
        print(f"connecting to '{channel.name}'")

    def disconnect_from_channel(self, channel):
        self._send_raw("PART #"+ channel.name.lower() +" ")
        self.__clearchannel(channel)
        print(f"departing from '{channel.name}'")

    def __clearchannel(self, channel: str):
        counter = 0
        for chn in self.channel_list:
            if chn.name == channel:
                self.channel_list.pop(counter)
                return
            counter += 1

    def activate(self, channels):
        self.channel_list = channels
        self.__connect()
        for channel in self.channel_list:
            self._send_raw("JOIN #"+ channel.name.lower() +" ")

    def run(self):
        while self.active:
            self.__process_stream_data()
            sleep(0.00001)
        try:
            self.socket.close()
        except Exception:
            pass

    def close(self):
        print(f"({self.name}) Closing Socket")
        self.active = False
        self.socket.close()
        


    def _send_raw(self, message: str):
        try:
            self.socket.send((f"{message}\r\n").encode('utf-8'))
            if message[:4] == "PASS":
                print(f"({self.name}) < PASS ****")
            else:
                print(f"({self.name}) < {message}")
        except OSError:
            print(f"Socket is closed and must be reopened to send the message '{message}'")

    def __process_stream_data(self):
        try:
            self.buffer = self.buffer + self.socket.recv(1024).decode()
        except ConnectionAbortedError:
            print(f"({self.name}) Socket Closed")
        temp = self.buffer.split("\n")
        self.buffer = temp.pop()
        for line in temp:
            # print(f"({self.name}) > {line}")
            if ("PING :tmi.twitch.tv" in line): # Keep Alive Mechanism
                self._send_raw("PONG :tmi.twitch.tv")
            self.on_socket_data(line)


    def on_socket_data(self, line: str):
        raise NotImplementedError("`on_socket_data` not implemented")






class ReadSocket(Socket):

    def __init__(self, nick: str, token: str):
        Socket.__init__(self, nick, token)
        self.name = "Thread-ReadSocket"


    def on_socket_data(self, line: str):
        self.on_data(parse_line(line))

    def on_data(self, data):
        # print(f" >", data.line)
        for channel in self.channel_list:
            channel.on_raw(data)
        
        if data.inner == "Message":
            print(f"Message: {data.message}")
            for channel in self.channel_list:
                if channel.name in data.line:
                    channel.on_message(data)
        elif data.inner == "Join":
            print("Join Detected")
            for channel in self.channel_list:
                if channel.name in data.line:
                    channel.on_join(data)
        elif data.inner == "PrivateMessage":
            for channel in self.channel_list:
                if channel.name == data.channel:
                    print (f"#{data.channel} - {data.display_name}: {data.message_text}")
                    channel.on_privmessage(data)
        elif data.inner == "CommandMessage":
            for channel in self.channel_list:
                if channel.name == data.channel:
                    print (f"#{data.channel} - {data.display_name}: [CMD] {data.message_text}")
                    channel.on_command(data) 
    

class SendSocket(Socket):

    def __init__(self, nick: str, token: str):
        Socket.__init__(self, nick, token)
        
        self.name = "Thread-SendSocket"

    def join(self, channel: str):
        send = f"JOIN #{channel.lower()}"
        self._send_raw(send)

    def send_message(self, channel: str, message: str):
        send = f"PRIVMSG #{channel.lower()} :{message}"
        self._send_raw(send)
    
    def send_action_message(self, channel: str, message: str):
        send = f"PRIVMSG #{channel.lower()} :/me {message}"
        self._send_raw(send)

    def send_whisper(self, user: str, message: str):
        send = f"PRIVMSG #{self.nick} :/w {user} {message}"
        self._send_raw(send)

    def timeout_user(self, user: str, channel: str, timeout=1):
        send = f"PRIVMSG #{channel} :/timeout {user} {timeout}"
        self._send_raw(send)

    def clear_message(self, channel: str, message_id: str):
        send = f"PRIVMSG #{channel} :/delete {message_id}"
        self._send_raw(send)

    def ban_user(self, user: str, channel: str):
        send = f"PRIVMSG #{channel} :/ban {user}"
        self._send_raw(send)

    def on_socket_data(self, line: str):
        pass


from ..command import Command
from ..message import PrivateMessage, CommandMessage

class Connections(Command):

    def __init__(self, channel):
        Command.__init__(self, "Connections")
        self.channel = channel
    
    def on_command(self, data: CommandMessage):
        if data.KEYWORD == "join":
            if data.channel != self.channel.name and \
            (data.args[1] == data.display_name.lower() or data.display_name == "cubbei"):
                self.channel.socket.connect_to_channel(data.args[1])
                self.channel.socket.send_message(data.channel, f"Attempting to connect to `{data.args[1]}`")
                log(f"Attempting to connect to `{data.args[1]}`")
            else:
                log(f"unauthorised attempt to connect to `{data.args[1]}`")

def log(msg):
    print(f"Connections: {msg}")


def setup(channel):
    channel.load_command(Connections(channel))
    log(f"[{channel.name}]: Loaded Module Connections")

def teardown(channel):
    log(f"[{channel.name}]: Removed Module Connections")